@@ -1,6 +1,7 @@
 import io
 import os
 import json
+from base64 import b64encode
 from pathlib import Path
 
 import pytest
@@ -11,38 +12,39 @@ from tests import util
 
 
 def test_check_config_order(mocker):
+    server_cas = b64encode(b"foo").decode()
     env_config = json.loads(
-        """
-      {
-        "v2": {
+        f"""
+      {{
+        "v2": {{
           "active_context": "local",
-          "contexts": {
-            "local": {
+          "contexts": {{
+            "local": {{
               "pachd_address": "grpcs://172.17.0.6:30650",
-              "server_cas": "foo",
+              "server_cas": "{server_cas}",
               "session_token": "bar",
               "active_transaction": "baz"
-            }
-          }
-        }
-      }
+            }}
+          }}
+        }}
+      }}
     """
     )
     spout_config = json.loads(
-        """
-      {
-        "v2": {
+        f"""
+      {{
+        "v2": {{
           "active_context": "local",
-          "contexts": {
-            "local": {
+          "contexts": {{
+            "local": {{
               "pachd_address": "[::1]:80",
-              "server_cas": "foo",
+              "server_cas": "{server_cas}",
               "session_token": "bar",
               "active_transaction": "baz"
-            }
-          }
-        }
-      }
+            }}
+          }}
+        }}
+      }}
     """
     )
     local_config = json.loads(
@@ -97,6 +99,9 @@ def test_check_config_order(mocker):
 
 
 def test_parse_config():
+    server_cas = b"foo"
+    server_cas_base64 = b64encode(server_cas)
+
     (
         host,
         port,
@@ -114,7 +119,7 @@ def test_parse_config():
             "contexts": {
               "local": {
                 "pachd_address": "grpcs://172.17.0.6:30650",
-                "server_cas": "foo",
+                "server_cas": "%s",
                 "session_token": "bar",
                 "active_transaction": "baz"
               }
@@ -122,13 +127,14 @@ def test_parse_config():
           }
         }
         """
+            % server_cas_base64.decode()
         )
     )
     assert host == "172.17.0.6"
     assert port == 30650
     assert pachd_address == "grpcs://172.17.0.6:30650"
     assert auth_token == "bar"
-    assert root_certs == b"foo"
+    assert root_certs == server_cas
     assert transaction_id == "baz"
     assert tls is True
 
@@ -271,22 +277,23 @@ def test_client_new_from_config():
         )
 
     # check that pachd address and other context fields are respected
+    server_cas = b64encode(b"foo").decode()
     client = python_pachyderm.Client.new_from_config(
         config_file=io.StringIO(
-            """
-        {
-          "v2": {
+            f"""
+        {{
+          "v2": {{
             "active_context": "local",
-            "contexts": {
-              "local": {
+            "contexts": {{
+              "local": {{
                 "pachd_address": "grpcs://172.17.0.6:30650",
-                "server_cas": "foo",
+                "server_cas": "{server_cas}",
                 "session_token": "bar",
                 "active_transaction": "baz"
-              }
-            }
-          }
-        }
+              }}
+            }}
+          }}
+        }}
     """
         )
     )
